#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
File: config.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/31 15:56:05
"""

import os.path
from typing import List
from datetime import datetime
from fatbulls.core.exception import FactorBackTestException
from fatbulls.factor.constant import UNIV_NAME, RET_TYPE


class BackTestConfig(object):

    def __init__(self, sd: str, ed: str, task_num: str, backtest_name: str):
        self.today = datetime.now().strftime("%Y%m%d")
        self.backtest_day = datetime.now().strftime("%Y%m%d")
        self.sd = sd
        self.ed = ed
        self.task_num = task_num
        self.backtest_name = backtest_name

    @property
    def sd(self):
        return self._sd

    @sd.setter
    def sd(self, value):
        if not isinstance(value, str) or value < '20130000':
            raise FactorBackTestException('invalid sd date inputs')
        self._sd = value

    @property
    def ed(self):
        return self._ed

    @ed.setter
    def ed(self, value):
        if not isinstance(value, str):
            raise FactorBackTestException('invalid ed date inputs')
        self._ed = value


class FactorBackTestConfig(BackTestConfig):

    def __init__(self, sd: str, ed: str, task_num: str = 'WaitForOne', backtest_name: str = 'test_backtest',
                 univ_name: str = 'cnall', fdays: int = 1, rettype: str = 'vwap30', factor_dirs: List = None):
        BackTestConfig.__init__(self, sd, ed, task_num, backtest_name)
        if factor_dirs is None:
            raise FactorBackTestException('factor directory path is empty')
        self.univ_name = univ_name
        self.fdays = fdays
        self.rettype = rettype
        self.factor_dirs = factor_dirs

    @property
    def univ_name(self):
        return self._univ_name

    @univ_name.setter
    def univ_name(self, value):
        if not isinstance(value, str) or value not in UNIV_NAME:
            raise FactorBackTestException("preset univ_name {} not found".format(value))
        self._univ_name = value

    @property
    def rettype(self):
        return self._rettype

    @rettype.setter
    def rettype(self, value):
        if not isinstance(value, str) or value not in RET_TYPE:
            raise FactorBackTestException("preset rettype {} not found".format(value))
        self._rettype = value

    @property
    def fdays(self):
        return self._fdays

    @fdays.setter
    def fdays(self, value):
        if not (isinstance(value, int) and 1 <= value <= 63):
            raise FactorBackTestException("forward looking days should fall into range 1~63 days")
        self._fdays = value

    @property
    def factor_dirs(self):
        return self._factor_dirs

    @factor_dirs.setter
    def factor_dirs(self, value):
        if not (isinstance(value, List) and len(value) >= 1):
            raise FactorBackTestException("must pass at least 1 valid factor path")
        for dirpath in value:
            if not os.path.exists(dirpath):
                raise FactorBackTestException("factor directory path not exists: {}".format(dirpath))
        self._factor_dirs = value


class PortfolioBacktestConfig(object):

    def __init__(self, config: dict):
        self.config = config








#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: engine.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/1 09:44:49
"""

from six import with_metaclass
from abc import ABCMeta, abstractmethod


class AbstractBackTestEngine(with_metaclass(ABCMeta)):

    @abstractmethod
    def run(self):
        raise NotImplementedError

    @abstractmethod
    def save_success_flag(self):
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: engine.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/1 09:44:49
"""

"""portfolio文件列名"""
PORTFOLIO_COLUMNS = ['K', 'D', 'T', 'V', 'W', 'entry_dt', 'entry_time', 'init_W', 'yest_W']
PORTFOLIO_MAX_WEIGHT = 0.95
PORTFOLIO_WEIGHT_FIELD = 'W'

"""回测模式：日回测"""
MATCHING_TYPE_DAY_MODE = [
    'current_bar',  # 对应c2c, 以当前bar为收盘价。
    'vwap',         # 成交量加权平均价撮合。
]

"""回测模式：分钟回测"""
MATCHING_TYPE_MIN_MODE = [
    'current_bar',  # 对应c2c, 以当前bar为收盘价。
    'vwap',         # 成交量加权平均价撮合。
    'next_bar',     # 对应o2o, 下一个 bar 的开盘价
]

"""回测模式：TICK回测"""
MATCHING_TYPE_TICK_MODE = [
    'last',                 # 以最新价
    'best_own',             # 己方最优价
    'best_counterparty',    # 对手方最优价撮合
    'counterparty_offer',   # 逐档撮合
]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


import pandas as pd
import numpy as np
import openpyxl
import traceback
from ..constant import UNIV_NAME, RET_TYPE, IND_DICT
import fqdatac
import duckdb
import os
from dataclasses import dataclass
from functools import reduce
from datetime import datetime

from fqdatac.etl.db_tools import timer


class FQDataLoader(object):
    def __init__(self):
        self.today = datetime.now().strftime("%Y%m%d")
        pd.set_option("display.max_columns", 100)
        pd.set_option("display.width", 1000)
        pd.set_option("display.max_colwidth", 100)
        pd.set_option("display.float_format", lambda x: "%.4f" % x)
        addr_prod = ("api.datavita.com.cn", 443)
        addr_dev = ("219.143.244.230", 8390)
        fqdatac.init(access_key="123", access_secret="456", addr=addr_dev, timeout=60)

    @staticmethod
    def partition_df(df: pd.DataFrame, col: str):
        """Partition a DataFrame into a dictionary of DataFrames by column value."""
        return {k: v.reset_index(drop=True) for k, v in df.groupby(col, sort=False)}


@dataclass
class Factor:
    """
    一个因子的全部数据类
    """

    hist: pd.DataFrame

    def __post_init__(self):
        assert set(["K", "D", "T"]) < set(self.hist.columns)
        self.facname = list(set(self.hist.columns) - set(["K", "D", "T"]))[0]
        self.hist = self.hist.sort_values(by=["D", "T"])
        self.hist_d = FQDataLoader.partition_df(self.hist, "D")
        self.facdays = sorted(list(set(self.hist["D"])))

    @property
    def coverage(self):
        return self.hist["K"].nunique()

    @property
    def sd(self):
        return self.facdays[0]

    @property
    def ed(self):
        return self.facdays[-1]

    @property
    def numdays(self):
        return len(fqdatac.get_daysrange(self.sd, self.ed))

    @property
    def hist_k(self):
        return FQDataLoader.partition_df(self.hist, "K")

    @classmethod
    def read_factor(cls, path, sd):
        factor_data = []
        for file in os.listdir(path):
            dt, fmt = file.split(".")
            if dt >= sd and fmt == "csv":
                factor_data.append(pd.read_csv(os.path.join(path, file)))
        alldata = pd.concat(factor_data, axis=0)
        return cls(hist=alldata)

    @classmethod
    def from_dict(cls, d: dict):
        factor_data = [v for v in d.values()]
        alldata = pd.concat(factor_data, axis=0)
        return cls(hist=alldata)

    def save_tocsv(self, path):
        os.makedirs(path, exist_ok=True)
        for k, v in self.hist_d.items():
            v.to_csv(os.path.join(path, f"{k}.csv"))


@dataclass
class FactorFile:
    """
    一个因子某单日的因子文件
    """

    kdtv: pd.DataFrame
    fac_name: str
    dt: str

    def __post_init__(self):
        assert set(["K", "D", "T"]) < set(self.kdtv.columns)


class FactorBacktest(FQDataLoader):
    def __init__(
            self, sd, ed, backtest_name, univ_name, fdays, rettype, directory_list
    ):
        self.backtest_day = datetime.now().strftime("%Y%m%d")
        self.sd = sd
        self.ed = ed
        self.backtest_name = backtest_name
        self.univ_name = univ_name
        self.fdays = fdays
        self.rettype = rettype
        self.directory_list = directory_list
        self.output_dir = "/data/quant/report/factor_backtest"

    def check_backtest_legitimacy(self):
        """
        check to see if legitimate backtest inputs are given
        """
        # univ_name必须在预设的list里面
        try:
            assert self.univ_name in UNIV_NAME
        except AssertionError:
            raise ValueError(f"preset univ_name {self.univ_name} not found")

        # ret_type必须在预设的list里面
        try:
            assert self.rettype in RET_TYPE
        except AssertionError:
            raise ValueError(f"preset ret_type {self.rettype} not found")

        # 向前预测期必须在1~63天的整数之内
        try:
            assert isinstance(self.fdays, int) and 1 <= self.fdays <= 63
        except AssertionError:
            raise ValueError("forward looking days should fall into range 1~63 days")

        # 必须至少传一个存在的factor路径
        try:
            assert len(self.directory_list) >= 1
        except AssertionError:
            raise ValueError("must pass at least 1 valid factor path")

    def checkdates(self):
        """
        查回测日期区间
        """
        try:
            assert isinstance(self.ed, str) and isinstance(self.sd, str)
            assert fqdatac.get_day(self.ed, -60) >= self.sd
            assert self.ed <= fqdatac.get_day(self.backtest_day, -1)
            assert self.sd >= "20130000"
        except AssertionError:
            raise KeyError("invalid date range inputs")

    def get_univ_cnall(self):
        univ = fqdatac.get_universe(sd=self.sd, ed=self.ed, univ_name="cnall")
        return self.partition_df(univ, "D")

    def get_univ_userselect(self):
        univ = fqdatac.get_universe(sd=self.sd, ed=self.ed, univ_name=self.univ_name)
        return self.partition_df(univ, "D")


class FactorBTLoader(FactorBacktest):

    """
    因子回测时的加载类
    """

    def __init__(
            self, sd, ed, backtest_name, univ_name, fdays, rettype, directory_list
    ):
        super(FactorBTLoader, self).__init__(
            sd, ed, backtest_name, univ_name, fdays, rettype, directory_list
        )
        self.check_backtest_legitimacy()
        self.checkdates()
        self.container = dict()
        self.alldays = fqdatac.get_daysrange(self.sd, self.ed)
        self.cnall = self.get_univ_cnall()
        self.univ = self.get_univ_userselect()

    def load(self, directory) -> dict:
        """
        将因子load进container
        """
        self.checker = FactorFileChecker(
            sd=self.sd,
            ed=self.ed,
            backtest_name=self.backtest_name,
            univ_name=self.univ_name,
            fdays=self.fdays,
            rettype=self.rettype,
            directory_list=self.directory_list,
            directory=directory,
        )

        files_to_load = self.checker.checkfiles()  # 查路径里的文件是否齐全 是否重复
        fullpath = [os.path.join(directory, x) for x in files_to_load]
        facdict = {}
        for f in fullpath:
            dtcsv = pd.read_csv(f)
            dtcsv["D"] = dtcsv["D"].astype(str)
            self.checker.checkformat(f, dtcsv)  # 查一下文件结构是否符合规范
            facdict[dtcsv["D"].iloc[0]] = dtcsv  # 确保一个文件里只有一个D，已经查过了
        self.container[self.checker.facname] = facdict

    def onlyuniv(self):
        """
        只保留当天在universe中的股票进行测试
        """
        for k, v in self.container.items():
            for dt in list(v.keys()):
                v[dt] = v[dt][v[dt]["K"].isin(list(self.univ[dt]["K"]))]

    def fwdret_load(self):
        """
        获取对照的股票return数据
        """
        self.fwdret1 = None
        self.fwdret = None
        self.fwdret = fqdatac.get_bfwd(
            sd=self.sd,
            ed=self.ed,
            univ_name=self.univ_name,
            ret_type=self.rettype,
            dayshift=self.fdays,
        )
        # self.fwdret = self.partition_df(fwdret, "D")
        if self.fdays == 1:
            self.fwdret1 = self.fwdret.copy()
        else:
            self.fwdret1 = fqdatac.get_bfwd(
                sd=self.sd,
                ed=self.ed,
                univ_name=self.univ_name,
                ret_type=self.rettype,
                dayshift=1,
            )
        #     self.fwdret1 = self.partition_df(fwdret1, "D")


class FactorFileChecker(FactorBTLoader):

    """
    因子路径、文件的格式校验
    """

    def __init__(
            self,
            sd,
            ed,
            backtest_name,
            univ_name,
            fdays,
            rettype,
            directory_list,
            directory,
    ):
        super(FactorFileChecker, self).__init__(
            sd, ed, backtest_name, univ_name, fdays, rettype, directory_list
        )
        self.path = directory
        self.facname = self.path.split("/")[-1]

    def checkfiles(self):
        """
        查重复文件，查缺失文件
        """
        files = os.listdir(self.path)
        csv_files = sorted([x for x in files if x.endswith(".csv")])

        files_alldays = [f"{x}.csv" for x in self.alldays]
        files_shouldbe = sorted([x for x in files_alldays if x >= csv_files[0]])
        files_existing = sorted([x for x in csv_files if x in files_alldays])

        # 查找重复文件
        counts = [files_existing.count(item) for item in files_existing]
        if max(counts) >= 2:
            raise FileExistsError("duplicate files of the same day")

        # 查找缺失文件
        missing_files = list(set(files_shouldbe) - set(files_existing))
        if len(missing_files) > 0:
            miss_days = ", ".join(missing_files)
            raise FileNotFoundError(f"the following files are missing: {miss_days}")

        return files_shouldbe

    def checkformat(self, f: str, df: pd.DataFrame):
        """
        查每个文件里的格式
        """

        # 查表头
        if not {"K", "D", "T", self.facname} == set(df.columns):
            raise KeyError(f"incorrect header format in file {f}")

        # 查文件内是否有重复stock code
        df_dd = df.drop_duplicates(subset=["K"])
        if len(df_dd) < len(df):
            raise ValueError(f"duplicate stock codes in file {f}")

        # 查文件内日期是否与文件名一致，是否有不同日期出现在一个文件
        dt = (f.split("/")[-1]).split(".")[0]
        try:
            assert max(df.D) == min(df.D) == dt
        except AssertionError:
            raise ValueError(
                f"multiple dates in same file/filename doesn't match D column in file {f}"
            )

        # 查文件内的股票代码是否都是上市的股票
        stk_code = set(df_dd["K"])
        if not set(stk_code) <= set(self.cnall[dt]["K"]):
            raise ValueError(f"stock code beyond cnall exists in file {f}")


class FactorEvaluator(FactorBacktest):
    """
    因子评估器，用于加载因子到评估器，评估并输出
    """

    def __init__(
            self,
            sd,
            ed,
            backtest_name,
            univ_name,
            fdays,
            rettype,
            directory_list,
            fwdret,
            fwdret1,
            raw: dict,
            output_path: str,
            task_num: str,
            conn=None,
    ):  # 这里的data应该是之前FactorLoader.container
        super(FactorEvaluator, self).__init__(
            sd,
            ed,
            backtest_name,
            univ_name,
            fdays,
            rettype,
            directory_list,
        )
        self.alldays = fqdatac.get_daysrange(self.sd, self.ed)
        self.raw = raw
        self.fwd = fwdret
        self.fwd1 = fwdret1
        self.conn = conn
        self.output_path = output_path
        self.task_num = task_num

    def initialize(self):
        self.fac_processing = False
        self.whole_processing = True
        self.iterfac = iter(self.raw.keys())
        self.all_report = None

    # 切换到下一个因子去评估，将整个因子的数据加载到容器
    def nextfactor(self):
        try:
            self.curfac = next(self.iterfac)
            self.curraw = Factor.from_dict(self.raw[self.curfac])
            self.fac_processing = not self.fac_processing
            # self.iterdays = iter(sorted(self.curraw.facdays))
        except StopIteration:
            self.whole_processing = not self.whole_processing

    # 将下一天的因子和fwdret加载到container
    def nextfile(self):
        try:
            dt = next(self.iterdays)
            self.ff = FactorFile(
                kdtv=self.curraw.hist_d[dt], fac_name=self.curfac, dt=dt
            )
            self.curfwdret = self.fwd[dt]
        except StopIteration:
            self.fac_processing = not self.fac_processing

    # 将前一天的因子文件存起来，后面可能会有用
    def save_lastfile(self):
        self.lastfile = self.ff

    # 评估因子文件
    @timer
    def evaluate(self):
        try:
            # assert self.fac_processing and self.whole_processing
            assert self.whole_processing
        except AssertionError:
            raise IndexError(
                "the whole_process status flag and factor status flag are not both true"
            )

        facind = FactorIndicator(self.curraw.hist, self.fwd, self.fwd1, self.sd, self.ed)
        self.ic = facind.ic()
        # self.hitrate= facind.hitrate() #
        self.hitrate, self.graph_hit_rate_by_group = facind.hitrate_re()  ##
        self.graph_hitrate = facind.graph_hitrate_re()  ##
        self.ret, self.ret_by_group = facind.ret()
        self.wealth, self.wealth_by_group = facind.wealth()
        self.graph_return = facind.grpah_ret()
        # self.turnover = facind.turnover() #
        self.turnover = facind.turnover_re()
        self.coverage = facind.coverage()
        self.serial_correlation = facind.serial_corr()
        self.summary = facind.get_summary()
        self.table_performance_by_group = facind.performace_by_group
        self.graph_coverage = facind.graph_coverage()
        self.graph_ic = facind.graph_ic()
        self.graph_ic_by_industry = facind.graph_ic_industry()
        # self.graph_hit_rate_by_group = facind.graph_hit_rate_10bin() #
        self.graph_turnover = facind.graph_turnover_monthly()
        # self.graph_hitrate = facind.graph_hitrate_monthly() ##
        self.curreport = self.sf_report()

    # 合并得到包含单因子所有评价信息的dict
    def sf_report(self):
        single_report = {
            "Summary": self.summary["LS"],  # long-short 组合的performance
            "Coverage": self.coverage,
            "ICs": self.ic,
            "Turnover": self.turnover["LS"],  ####
            "graph_Turnover": self.graph_turnover["LS"],
            "LSreturns": self.graph_return["LS"],
            "Wealth": self.wealth["LS"],
            "HitRate": self.hitrate["LS"],  ####
            "graph_HitRate": self.graph_hitrate["LS"],
            "SCs": self.serial_correlation,
            "top_Summary": self.summary["top"],  # top 组合的performance
            "top_Turnover": self.turnover["top"],  ####
            "graph_top_Turnover": self.graph_turnover["top"],
            "top_Returns": self.graph_return["top"],
            "top_Wealth": self.wealth["top"],
            "top_HitRate": self.hitrate["top"],  ####
            "graph_top_HitRate": self.graph_hitrate["top"],
            "bottom_Summary": self.summary["bottom"],  # bottom 组合的performance
            "bottom_Turnover": self.turnover["bottom"],  ####
            "graph_bottom_Turnover": self.graph_turnover["bottom"],
            "bottom_Returns": self.graph_return["bottom"],
            "bottom_Wealth": self.wealth["bottom"],
            "bottom_HitRate": self.hitrate["bottom"],  ####
            "graph_bottom_HitRate": self.graph_hitrate["bottom"],
            # 图表的内容
            "graph_coverage": self.graph_coverage,
            "graph_ic": self.graph_ic,  # IC 月度图，key：D
            "graph_ic_by_industry": self.graph_ic_by_industry,  # IC分行业图, key:month
            "graph_wealth_by_group": self.wealth_by_group,  # 因子分10组净值曲线图，key：bin，D
            "graph_ret_by_group": self.ret_by_group,  #### 因子分10组收益率数据，key：bin，D
            "graph_hit_rate_by_group": self.graph_hit_rate_by_group,  # 分10组胜率图，key：bin
            "table_performance_by_group": self.table_performance_by_group,  # 分10组收益率，夏普比率，波动率，最大回撤，
        }
        return single_report

    def fac_report(self):
        # 将所有因子的report合并到总的report里面
        if self.all_report == None:
            self.all_report = {key: [] for key in self.curreport}
            for key in self.curreport:
                self.all_report[key].append(self.curreport[key])
        else:
            for key in self.curreport:
                self.all_report[key].append(self.curreport[key])

    # 将评估完的因子进行输出
    def output(self):
        # 对sel.all_report 遍历key，将结果dataframe concat起来，得到所有因子的report
        dict_output = {}
        dict_output["corr_matrix"] = self.factor_corrmatrix().reset_index()
        for key in self.all_report:
            try:
                if key in [
                    "Summary",
                    "top_Summary",
                    "bottom_Summary",
                    "table_performance_by_group",
                ]:
                    dict_output[key] = pd.concat(self.all_report[key], axis=0)
                    dict_output[key] = dict_output[key].reset_index(drop=True)
                elif key in ["graph_wealth_by_group", "graph_ret_by_group"]:
                    dict_output[key] = reduce(
                        lambda x, y: pd.merge(x, y, on=["D", "bin"], how="outer"),
                        self.all_report[key],
                    )
                    dict_output[key] = dict_output[key].reset_index(drop=True)
                elif key in ["graph_hit_rate_by_group"]:
                    dict_output[key] = reduce(
                        lambda x, y: pd.merge(x, y, on=["bin"], how="outer"),
                        self.all_report[key],
                    )
                    dict_output[key] = dict_output[key].reset_index(drop=True)
                elif key in ["graph_ic_by_industry"]:
                    dict_output[key] = reduce(
                        lambda x, y: pd.merge(
                            x, y, on=["ind_code", "ind_name", "bin"], how="outer"
                        ),
                        self.all_report[key],
                    )
                    dict_output[key] = dict_output[key].reset_index(drop=True)
                else:
                    dict_output[key] = reduce(
                        lambda x, y: pd.merge(x, y, on=["D"], how="outer"),
                        self.all_report[key],
                    )
                    dict_output[key] = dict_output[key].reset_index(drop=True)
            except:
                traceback.print_exc()
                print(f"多因子报告合并: {key}合并失败")
        # 写到Excel
        wb = openpyxl.Workbook(write_only=True)
        for sheet_name, df in dict_output.items():
            try:
                if isinstance(df, pd.DataFrame):
                    ws = wb.create_sheet(title=sheet_name)
                    # 写入header
                    ws.append(df.columns.tolist())
                    # 写入每行数据
                    for row in df.values.tolist():
                        ws.append(row)
            except:
                traceback.print_exc()
                print(f"excel writer: {sheet_name} 写入失败")

        today = datetime.now().strftime("%Y%m%d")
        file_name = f"BT_factor_report_{today}_{self.task_num}_{self.univ_name}_{self.sd}_{self.ed}.xlsx"

        # # 输出到指定本地路径
        # local_dir = self.output_path + "/" + self.task_num
        # os.makedirs(local_dir, exist_ok=True)
        # wb.save(local_dir + "/" + file_name)

        # # 生成程序运行成功信号
        # try:
        #     wb = openpyxl.load_workbook(local_dir + "/" + file_name)
        # except Exception as e:
        #     raise Exception(f"The generated Excel file could not be read. Reason: {e}")

        # success_file = os.path.join(local_dir, "success.txt")
        # with open(success_file, "w") as f:
        #     print(f"Generated {success_file}")    

        # 输出到指定路径
        try:
            output_dir = self.output_dir + "/" + self.task_num
            os.makedirs(output_dir, exist_ok=True)
            wb.save(output_dir + "/" + file_name)
            # 生成程序运行成功信号
            try:
                wb = openpyxl.load_workbook(output_dir + "/" + file_name)
            except Exception as e:
                raise Exception(f"The generated Excel file could not be read. Reason: {e}")
            success_file = os.path.join(output_dir, "success.txt")
            with open(success_file, "w") as f:
                print(f"Generated {success_file}")
        except:
            traceback.print_exc()


    # 因子相关性
    @timer
    def factor_corrmatrix(self):
        factor_names = list(self.raw.keys())
        dates = list(self.raw[factor_names[0]].keys())
        corr_matrix = pd.DataFrame(index=factor_names, columns=factor_names)
        for i in range(len(factor_names)):
            for j in range(i + 1, len(factor_names)):
                # 存储每日相关系数
                daily_corrs = []
                for d in dates:
                    # 如果对应日期的数据不存在，则返回nan
                    if d not in self.raw[factor_names[i]].keys() or d not in self.raw[factor_names[j]].keys():
                        daily_corrs.append(np.nan)
                        continue
                    fac1 = self.raw[factor_names[i]][d]
                    fac2 = self.raw[factor_names[j]][d]
                    # 保留具有相同股票的数据
                    merged = pd.merge(fac1, fac2, on=["K", "D", "T"], how="inner")
                    # 获取两个因子的列名
                    fac1_name = list(fac1.columns)[-1]
                    fac2_name = list(fac2.columns)[-1]
                    # 去除空值
                    merged.dropna(subset=[fac1_name, fac2_name], inplace=True)
                    # 计算每日相关系数
                    daily_corr = merged[fac1_name].corr(merged[fac2_name])
                    # 添加到列表
                    daily_corrs.append(daily_corr)
                # 求列表的平均值
                mean_corr = pd.Series(daily_corrs).mean().round(4)
                # 填入矩阵
                corr_matrix.loc[factor_names[i], factor_names[j]] = mean_corr
                corr_matrix.loc[factor_names[j], factor_names[i]] = mean_corr
            corr_matrix.loc[factor_names[i], factor_names[i]] = 1.0000

        return corr_matrix


class FactorIndicator:
    """
    factor indicators and calculation
    """

    def __init__(self, fac, fwd, fwd1, sd, ed):
        self.fac = fac
        self.fwd = fwd
        self.fwd1 = fwd1
        self.sd = sd
        self.ed = ed
        self.fac_name = list(self.fac.columns)[-1]
        self.fwd_ret_name = list(self.fwd.columns)[-1]
        self.fwd_ret_name1 = list(self.fwd1.columns)[-1]
        self.fac_fwd = self.fac_name + "_" + self.fwd_ret_name
        # 获取画图的参数：
        # 回测期一年以内为周，平均值rolling window 为4，
        # 超过一年为月，平均值rolling window 为6
        self.freq, self.window = self.get_graph_params()
        # self.fac_bin = FactorIndicator.bin_division(self.fac.copy(), 10)
        self.fac_bin = FactorIndicator.bin_division_re(self.fac.copy(), 10)

    # 获取画图groupby的参数
    def get_graph_params(self):
        d1 = pd.Period(self.sd + "M")
        d2 = pd.Period(self.ed + "M")
        n_month = pd.Series(pd.period_range(d1, d2, freq="M")).nunique()
        if n_month > 12:
            freq = "M"
            window = 6
        else:
            freq = "W-FRI"
            window = 4
        return freq, window

    # 判断本月有多少天
    def days_in_month(self, year, month):
        if month == 12:
            next_year = year + 1
            next_month = 1
        else:
            next_year = year
            next_month = month + 1
        end_of_month = datetime(next_year, next_month, 1) - datetime(year, month, 1)
        return (end_of_month).days

    # 判断日期间隔
    def get_date_interval(self, dt_start, dt_end):
        date_format = "%Y%m%d"

        # 将字符串日期转换为datetime对象
        start_date = datetime.strptime(dt_start, date_format)
        end_date = datetime.strptime(dt_end, date_format)

        # 计算日期间隔
        delta = end_date - start_date

        # 判断间隔
        if delta.days < self.days_in_month(start_date.year, start_date.month):
            return "day"
        elif delta.days < 365:
            return "week"
        else:
            return "month"

    @timer
    @staticmethod
    def bin_division(df: pd.DataFrame, nbins: int = 10):
        """
        将因子值分为x个bin
        结果的bin列范围为1~nbins，bin=1代表因子值最小组，bin=nbins代表因子值最大组
        """
        value_col = list(df.columns)[-1]
        df_dropna = df.copy().dropna(subset=[value_col])
        grouped = df_dropna.groupby(["D"])
        result_df = pd.DataFrame()
        for name, sub_df in grouped:
            sorted_df = sub_df.sort_values(by=value_col, ascending=True)
            sorted_df["rank"] = sorted_df.reset_index(drop=True).index + 1
            max_rank = max(sorted_df["rank"])
            sorted_df["rank_div_max"] = sorted_df["rank"] / max_rank
            try:
                assert nbins >= 1
                assert nbins <= max_rank
            except AssertionError:
                raise ValueError(
                    f"The range of nbins needs to be [1, length of the df]"
                )
            sorted_df["bin"] = ""
            for i in range(1, nbins + 1):
                bin_min = (i - 1) / float(nbins)
                bin_max = i / float(nbins)
                # 使用 .loc 来定位需要赋值的位置
                sorted_df.loc[
                    (sorted_df["rank_div_max"] > bin_min)
                    & (sorted_df["rank_div_max"] <= bin_max),
                    "bin",
                ] = i
            result_df = pd.concat([sorted_df, result_df])

        result_df.sort_values(by=["D", "bin"], ascending=True, inplace=True)
        return result_df[list(df.columns) + ["bin"]]

    @timer
    @staticmethod
    def bin_division_re(df: pd.DataFrame, nbins: int = 10):
        """
        将因子值分为x个bin
        结果的bin列范围为1~nbins，bin=1代表因子值最小组，bin=nbins代表因子值最大组
        """
        value_col = list(df.columns)[-1]
        result_df = df.dropna(subset=[value_col])
        result_df['rank'] = result_df.groupby('D')[value_col].rank(method='first')
        result_df['bin'] = result_df.groupby('D')['rank'].transform(lambda x: pd.qcut(x, nbins, labels=range(1,nbins+1,1)))
        result_df['bin'] = result_df['bin'].astype('int')
        result_df = result_df.drop(columns=["rank"])
        return result_df


    @timer
    @staticmethod
    def bin_division_duckdb(df: pd.DataFrame, nbins: int = 10):
        """
        将因子值分为x个bin
        结果的bin列范围为1~nbins，bin=1代表因子值最小组，bin=nbins代表因子值最大组
        """
        value_col = list(df.columns)[-1]
        df_dropna = df.dropna(subset=[value_col])
        con = duckdb.connect()
        con.register('df', df_dropna)
        sql = f"""
                WITH daily_ranked AS (
                SELECT 
                    *,
                    ROW_NUMBER() OVER (PARTITION BY "D" ORDER BY {value_col}) AS daily_rank 
                FROM df
                ),

                final AS (
                SELECT 
                    *,  COUNT(*) OVER (PARTITION BY "D") AS total_rows
                FROM daily_ranked 
                )

                SELECT
                *,
                CEIL(daily_rank * {nbins} / total_rows) AS bin
                FROM final
                """
        result_df = con.execute(sql).fetchdf()
        con.close()
        result_df['bin'] = result_df['bin'].astype(int)
        result_df.sort_values(by=["D", "bin"], ascending=True, inplace=True)
        return result_df[list(df.columns) + ["bin"]]


    # 定义一个resample函数，将时间序列按照给定频率的交易日末进行转换
    @staticmethod
    def df_resample(df, freq, method):
        # freq = "M","W-FRI"
        df.index = pd.to_datetime(df["D"], format="%Y%m%d")
        D_K = sorted(list(set(df.columns) & set(["D", "K"])))
        s_date = df.resample(freq).last()[D_K]
        if method == "mean":
            df = df.resample(freq).mean().reset_index(drop=True)
        elif method == "last":
            df = df.resample(freq).last().reset_index(drop=True)
        else:
            pass
        df[D_K] = s_date.values
        return df

    # 计算IC
    @timer
    def ic(self):
        fac = self.fac.dropna(subset=[self.fac_name])
        fwd = self.fwd.dropna(subset=[self.fwd_ret_name])
        merged = pd.merge(fac, fwd, how="inner", on=["K", "D"])

        def ic_group(group: pd.DataFrame) -> float:
            """
            计算T日的ic值
            """
            fac_rank = group[self.fac_name].rank(method="first")
            fwd_rank = group[self.fwd_ret_name].rank(method="first")
            rank_diff = fac_rank - fwd_rank
            n = len(group)
            if n <= 1:
                res = np.nan
            else:
                res = 1 - 6 * sum(rank_diff * rank_diff) / n / (n * n - 1)
            return res

        # 计算每天的IC值
        df_ic = merged.groupby("D").apply(ic_group).reset_index()
        df_ic.columns = ["D", self.fac_fwd]
        self.df_ic = df_ic
        return self.df_ic

    # 计算胜率
    @timer
    def hitrate(self):
        # 分组
        fac_bin = self.fac_bin.copy()
        # 合并因子值和收益率
        df_merge = pd.merge(fac_bin, self.fwd, on=["K", "D"], how="left")
        # 统计rk>0,rs<0,rk,rs,rk+rs的个数
        df_result = (
            df_merge.groupby("D")
            .agg(
                rs_0_cnt=(
                    self.fwd_ret_name,
                    lambda x: ((df_merge["bin"] == 1) & (x < 0)).sum(),
                ),
                rk_0_cnt=(
                    self.fwd_ret_name,
                    lambda x: ((df_merge["bin"] == 10) & (x > 0)).sum(),
                ),
                rs_cnt=("bin", lambda x: (x == 1).sum()),
                rk_cnt=("bin", lambda x: (x == 10).sum()),
                rk_rs_cnt=("bin", lambda x: ((x == 1) | (x == 10)).sum()),
            )
            .reset_index()
        )
        # 计算 hit rate
        df_result["LS"] = (df_result["rs_0_cnt"] + df_result["rk_0_cnt"]) / df_result[
            "rk_rs_cnt"
        ]
        df_result["top"] = df_result["rk_0_cnt"] / df_result["rk_cnt"]
        df_result["bottom"] = df_result["rs_0_cnt"] / df_result["rs_cnt"]
        # 构建结果字典
        self.dict_result_hitrate = {
            "LS": df_result[["D", "LS"]].rename(columns={"LS": self.fac_fwd}),
            "top": df_result[["D", "top"]].rename(columns={"top": self.fac_fwd}),
            "bottom": df_result[["D", "bottom"]].rename(
                columns={"bottom": self.fac_fwd}
            ),
        }

        return self.dict_result_hitrate

    @timer
    def hitrate_re(self):
        # 分组
        fac_bin = self.fac_bin.copy()
        # 合并因子值和收益率
        df_merge = pd.merge(fac_bin, self.fwd, on=["K", "D"], how="left")
        df_merge[self.fwd_ret_name] = df_merge[self.fwd_ret_name] > 0
        df_result = df_merge.groupby(by=["bin", "D"]).sum()[[self.fwd_ret_name]]
        df_result.columns = ["win"]
        df_result["count"] = df_merge.groupby(by=["bin", "D"]).count()[
            [self.fwd_ret_name]
        ]
        df_result["lose"] = df_result["count"] - df_result["win"]
        df_result = df_result.reset_index()
        # top 组合的hitrate
        top_hitrate = df_result[df_result["bin"] == 10].copy()
        top_hitrate[self.fac_fwd] = (
                top_hitrate.loc[:, "win"] / top_hitrate.loc[:, "count"]
        )
        # bottom 组合的hitrate
        bottom_hitrate = df_result[df_result["bin"] == 1].copy()
        bottom_hitrate.loc[:, self.fac_fwd] = (
                bottom_hitrate.loc[:, "lose"] / bottom_hitrate.loc[:, "count"]
        )
        # longshort 组合的hitrate
        top_df = top_hitrate[["D", "win", "lose", "count"]]
        top_df.columns = ["D", "win_top", "lose_top", "count_top"]
        bottom_df = bottom_hitrate[["D", "win", "lose", "count"]]
        bottom_df.columns = ["D", "win_bottom", "lose_bottom", "count_bottom"]
        ls_hitrate = pd.merge(top_df, bottom_df, how="left", on="D")
        ls_hitrate["ls_win"] = ls_hitrate["win_top"] + ls_hitrate["lose_bottom"]
        ls_hitrate["ls_count"] = ls_hitrate["count_top"] + ls_hitrate["count_bottom"]
        ls_hitrate.loc[:, self.fac_fwd] = (
                ls_hitrate.loc[:, "ls_win"] / ls_hitrate.loc[:, "ls_count"]
        )
        # 组合结果到dict
        columns = ["D", self.fac_fwd]
        ls_hitrate = ls_hitrate[columns].reset_index(drop=True)
        top_hitrate = top_hitrate[columns].reset_index(drop=True)
        bottom_hitrate = bottom_hitrate[columns].reset_index(drop=True)
        self.dict_result_hitrate = {
            "LS": ls_hitrate,
            "top": top_hitrate,
            "bottom": bottom_hitrate,
        }
        # 分10组的胜率
        df_result.loc[:, self.fac_fwd] = (
                df_result.loc[:, "win"] / df_result.loc[:, "count"]
        )
        hit_rate_by_group = (
            df_result.groupby(by="bin").mean()[[self.fac_fwd]].reset_index()
        )
        return self.dict_result_hitrate, hit_rate_by_group

    @timer
    def graph_hitrate_re(self):
        dict_hitrate = self.dict_result_hitrate.copy()
        # 计算到对应频率
        ls_hitrate = self.df_resample(dict_hitrate["LS"].copy(), self.freq, "mean")
        top_hitrate = self.df_resample(dict_hitrate["top"].copy(), self.freq, "mean")
        bottom_hitrate = self.df_resample(
            dict_hitrate["bottom"].copy(), self.freq, "mean"
        )
        # 计算平均线
        ls_hitrate[f"{self.fac_fwd}_avg"] = (
            ls_hitrate[self.fac_fwd].rolling(self.window).mean()
        )
        top_hitrate[f"{self.fac_fwd}_avg"] = (
            top_hitrate[self.fac_fwd].rolling(self.window).mean()
        )
        bottom_hitrate[f"{self.fac_fwd}_avg"] = (
            bottom_hitrate[self.fac_fwd].rolling(self.window).mean()
        )
        return {"LS": ls_hitrate, "top": top_hitrate, "bottom": bottom_hitrate}

    # 计算分组日度收益率
    @timer
    def ret(self):
        df_fac = self.fac_bin.copy()
        df_merge = pd.merge(df_fac, self.fwd1, on=["K", "D"], how="left")
        df_merge = df_merge.rename(columns={self.fwd_ret_name1: self.fac_fwd})
        # 计算分组收益率：日度
        self.ret_all = (
            df_merge[["D", "bin", self.fac_fwd]]
            .groupby(by=["D", "bin"])
            .mean()[self.fac_fwd]
            .reset_index()
        )
        self.ret_all = pd.pivot_table(
            self.ret_all, index="D", columns="bin", values=self.fac_fwd
        )
        self.ret_all = self.ret_all.shift(1).fillna(0).unstack().reset_index()
        self.ret_all.columns = ["bin", "D", self.fac_fwd]
        # 第1组的股票收益率：日度
        self.rs = self.ret_all[self.ret_all["bin"] == 1][["D", self.fac_fwd]]
        # 第10组的股票收益率：日度
        self.rk = self.ret_all[self.ret_all["bin"] == 10][["D", self.fac_fwd]]
        # Long Short 组合股票收益率：日度
        self.LSreturn = (
                self.rk.set_index("D", drop=True) - self.rs.set_index("D", drop=True)
        ).reset_index()
        return {"LS": self.LSreturn, "top": self.rk, "bottom": self.rs}, self.ret_all

    # 计算分组净值图，LS, top， bottom的净值图
    @timer
    def wealth(self):
        # 计算所有组合的净值
        self.wealth_by_group = self.ret_all.sort_values(by="D", ascending=True).copy()
        self.wealth_by_group = pd.pivot_table(
            self.wealth_by_group, columns="bin", index="D", values=self.fac_fwd
        )
        self.wealth_by_group = (
            (self.wealth_by_group + 1).cumprod().unstack().reset_index()
        )
        self.wealth_by_group.columns = ["bin", "D", self.fac_fwd]
        # 计算LS组合的净值
        self.df_wealth = self.LSreturn.copy()
        self.df_wealth[self.fac_fwd] = (self.df_wealth[self.fac_fwd] + 1).cumprod()
        # 计算top组合的净值
        self.top_wealth = self.rk.copy()
        self.top_wealth[self.fac_fwd] = (self.top_wealth[self.fac_fwd] + 1).cumprod()
        # 计算bottom组合的净值
        self.bottom_wealth = self.rs.copy()
        self.bottom_wealth[self.fac_fwd] = (
                self.bottom_wealth[self.fac_fwd] + 1
        ).cumprod()
        return {
            "LS": self.df_wealth,
            "top": self.top_wealth,
            "bottom": self.bottom_wealth,
        }, self.wealth_by_group

    # 计算单个组合的因子收益率图：周/月
    def sub_graph_ret(self, wealth, freq, method):
        # 转化净值为目标输出频率
        df_ret_freq = self.df_resample(wealth.copy(), freq, method)
        df_ret_freq = pd.concat([wealth.iloc[[0], :], df_ret_freq], axis=0).reset_index(
            drop=True
        )
        # 计算目标输出频率的因子收益率
        df_ret_freq[self.fac_fwd] = df_ret_freq[self.fac_fwd].pct_change()
        df_ret_freq = df_ret_freq.dropna()
        df_ret_freq[f"{self.fac_fwd}_avg"] = (
            df_ret_freq[self.fac_fwd].rolling(self.window).mean()
        )
        return df_ret_freq

    # 计算LS, top， bottom的因子收益率图:周/月
    @timer
    def grpah_ret(self):
        df_ret_freq = self.sub_graph_ret(self.df_wealth.copy(), self.freq, "last")
        top_ret_freq = self.sub_graph_ret(self.top_wealth.copy(), self.freq, "last")
        bottom_ret_freq = self.sub_graph_ret(
            self.bottom_wealth.copy(), self.freq, "last"
        )
        return {"LS": df_ret_freq, "top": top_ret_freq, "bottom": bottom_ret_freq}

    # 计算换手、序列相关性
    @timer
    def turnover(self):
        # 分成10组
        fac_bin = self.fac_bin.copy()
        # 获取日期列表
        dt_list = list(sorted(set(fac_bin["D"])))

        result_LS, result_top, result_bottom = [], [], []
        prev_s1_set, prev_s10_set = set(), set()

        for dt in dt_list:
            # 如果为第一天，换手率为1
            if dt == dt_list[0]:
                result_LS.append([dt, 1])
                result_top.append([dt, 1])
                result_bottom.append([dt, 1])
                # 第一日的股票集合，给第二天用
                prev_s1_set = set(
                    fac_bin[(fac_bin["D"] == dt) & (fac_bin["bin"] == 1)]["K"]
                )
                prev_s10_set = set(
                    fac_bin[(fac_bin["D"] == dt) & (fac_bin["bin"] == 10)]["K"]
                )
                continue

            dt_1 = fqdatac.get_day(dt, -1)

            # 当前日期的股票集合
            s1_set = set(fac_bin[(fac_bin["D"] == dt) & (fac_bin["bin"] == 1)]["K"])
            s10_set = set(fac_bin[(fac_bin["D"] == dt) & (fac_bin["bin"] == 10)]["K"])

            # T-1日的股票集合
            f1_set = prev_s1_set
            c1 = len(s1_set & f1_set)
            f10_set = prev_s10_set
            c10 = len(s10_set & f10_set)

            prev_s1_set, prev_s10_set = s1_set, s10_set

            if s1_set or s10_set:
                result_LS.append([dt, 1 - (c1 + c10) / (len(s1_set) + len(s10_set))])
                result_top.append([dt, 1 - c10 / len(s10_set)])
                result_bottom.append([dt, 1 - c1 / len(s1_set)])
            else:
                result_LS.append([dt, 0])
                result_top.append([dt, 0])
                result_bottom.append([dt, 0])

        df_LS = pd.DataFrame(result_LS, columns=["D", self.fac_fwd])
        df_top = pd.DataFrame(result_top, columns=["D", self.fac_fwd])
        df_bottom = pd.DataFrame(result_bottom, columns=["D", self.fac_fwd])
        self.dict_turnover = {"LS": df_LS, "top": df_top, "bottom": df_bottom}
        return self.dict_turnover

    @timer
    def turnover_re(self):
        # 分成10组
        fac_bin = self.fac_bin.copy()
        df_turnover = (
            fac_bin.groupby(by=["bin", "D"])["K"]
            .apply(lambda x: set(x.tolist()))
            .to_frame()
        )
        df_turnover.columns = ["K_set"]
        df_turnover = df_turnover.reset_index()
        df_turnover["K_count"] = df_turnover["K_set"].apply(lambda x: len(x))
        df_turnover["prev_K_set"] = (
            df_turnover.groupby("bin")["K_set"].apply(lambda x: x.shift(1)).values
        )
        df_turnover["overlap"] = df_turnover.apply(
            lambda x: x["K_set"] & x["prev_K_set"]
            if isinstance(x["K_set"], set) and isinstance(x["prev_K_set"], set)
            else 0,
            axis=1,
        )
        df_turnover["n_overlap"] = df_turnover["overlap"].apply(
            lambda x: len(x) if isinstance(x, set) else x
        )
        df_turnover[self.fac_fwd] = (
                1 - df_turnover["n_overlap"] / df_turnover["K_count"]
        )
        # 分10组的turnover
        self.turnover_by_group = df_turnover[["bin", "D", self.fac_fwd]]
        # top和bottom组合的turnover
        top_turnover = df_turnover[df_turnover["bin"] == 10][["D", self.fac_fwd]]
        bottom_turnover = df_turnover[df_turnover["bin"] == 1][["D", self.fac_fwd]]
        # longshort 组合的turnover
        cols = ["K_set", "K_count", "prev_K_set", "overlap", "n_overlap"]
        top_df = df_turnover[df_turnover["bin"] == 10][["D"] + cols]
        top_df.columns = ["D"] + [i + "_top" for i in cols]
        bottom_df = df_turnover[df_turnover["bin"] == 1][["D"] + cols]
        bottom_df.columns = ["D"] + [i + "_bottom" for i in cols]
        ls_turnover = pd.merge(top_df, bottom_df, how="left", on="D")
        ls_turnover["t_overlap"] = (
                ls_turnover["n_overlap_top"] + ls_turnover["n_overlap_bottom"]
        )
        ls_turnover["t_count"] = (
                ls_turnover["K_count_top"] + ls_turnover["K_count_bottom"]
        )
        ls_turnover[self.fac_fwd] = (
                1 - ls_turnover["t_overlap"] / ls_turnover["t_count"]
        )
        # self.dict_df_turnover = {"LS":ls_turnover.copy(),
        #                          'top':df_turnover[df_turnover["bin"] == 10],
        #                          'bottom':df_turnover[df_turnover["bin"] == 1]}
        ls_turnover = ls_turnover[["D", self.fac_fwd]]
        self.dict_turnover = {
            "LS": ls_turnover,
            "top": top_turnover,
            "bottom": bottom_turnover,
        }
        return self.dict_turnover

    # 计算覆盖率
    @timer
    def coverage(self):
        df = self.fac.dropna(subset=[self.fac_name])
        df_coverage = df.groupby("D")["K"].count().reset_index()
        df_coverage.columns = ["D", self.fac_fwd]
        return df_coverage

    # 计算因子序列相关性
    @timer
    def serial_corr(self):
        temp_df = self.fac.sort_values(by="D", ascending=True).copy()
        # 获取前一天的因子序列
        prev_fac = pd.pivot_table(
            temp_df, columns="K", index="D", values=self.fac_name
        ).shift(1)
        prev_fac = prev_fac.unstack().reset_index()
        prev_fac.columns = ["K", "D", "prev_fac"]
        prev_fac = prev_fac.dropna(subset=["prev_fac"])
        # 滚动计算相关性
        temp_df = pd.merge(temp_df, prev_fac, how="left", on=["K", "D"])
        self.serial_correlation = temp_df.groupby("D").apply(
            lambda x: x[[self.fac_name, "prev_fac"]].corr(method="spearman").iloc[0, 1]
        )
        self.serial_correlation = (
            self.serial_correlation.to_frame().reset_index().dropna()
        )
        self.serial_correlation.columns = ["D", self.fac_fwd]
        return self.serial_correlation

    # 计算summary中：coverage_max coverage_min
    @timer
    def summary_coverage(self):
        fac_dropna = self.fac.dropna(subset=[self.fac_name])
        coverage_max = len(fac_dropna["K"].unique())
        total_days = len(self.fac["D"].unique())
        df_days = fac_dropna.groupby("K")["D"].count()
        coverage_min = len(df_days[df_days == total_days])
        df_coverage = pd.DataFrame(
            {"coverage_max": [coverage_max], "coverage_min": [coverage_min]},
            index=[self.fac_fwd],
        )
        return df_coverage

    # 计算summary中：rank_ic_mean rank_ic_std ic_ir
    @timer
    def summary_ic_ir(self):
        df_ic = self.df_ic.copy()
        ic_mean = df_ic[self.fac_fwd].describe()["mean"] * 100
        rank_ic_mean = round(ic_mean, 2)
        ic_std = df_ic[self.fac_fwd].describe()["std"] * 100
        rank_ic_std = round(ic_std, 2)
        ic_ir = ic_mean / ic_std
        ic_ir = round(ic_ir, 2)
        df_ic_ir = pd.DataFrame(
            {
                "rank_ic_mean": [rank_ic_mean],
                "rank_ic_std": [rank_ic_std],
                "ic_ir": [ic_ir],
            },
            index=[self.fac_fwd],
        )
        return df_ic_ir

    # 计算summary中：rank_ic_mean_year_YYYY
    @timer
    def summary_ic_year(self):
        df_ic = self.df_ic.copy()
        df_ic["year"] = df_ic["D"].str[:4].astype(int)
        df_ic = df_ic.groupby("year")[self.fac_fwd].mean() * 100
        df_ic = round(df_ic, 2).to_frame().T
        df_ic_year = df_ic.add_prefix("rank_ic_mean_year_")
        return df_ic_year

    # 返回LS、top、bottom三组summary_hit_rate的dict，包含3个df
    @timer
    def summary_hit_rate(self):
        dict_hitrate = self.dict_result_hitrate
        hitrate_LS = pd.DataFrame(
            [round(dict_hitrate["LS"][self.fac_fwd].mean() * 100, 2)],
            columns=["hit_rate"],
            index=[self.fac_fwd],
        )
        hitrate_top = pd.DataFrame(
            [round(dict_hitrate["top"][self.fac_fwd].mean() * 100, 2)],
            columns=["hit_rate"],
            index=[self.fac_fwd],
        )
        hitrate_bottom = pd.DataFrame(
            [round(dict_hitrate["bottom"][self.fac_fwd].mean() * 100, 2)],
            columns=["hit_rate"],
            index=[self.fac_fwd],
        )
        return {"LS": hitrate_LS, "top": hitrate_top, "bottom": hitrate_bottom}

    # 返回LS、top、bottom三组summary_turnover的dict，包含3个df
    @timer
    def summary_turnover(self):
        dict_turnover = self.dict_turnover
        turnover_LS = pd.DataFrame(
            [round(dict_turnover["LS"][self.fac_fwd].mean() * 100,2)],
            columns=["turnover"],
            index=[self.fac_fwd],
        )
        turnover_top = pd.DataFrame(
            [round(dict_turnover["top"][self.fac_fwd].mean() * 100,2)],
            columns=["turnover"],
            index=[self.fac_fwd],
        )
        turnover_bottom = pd.DataFrame(
            [round(dict_turnover["bottom"][self.fac_fwd].mean() * 100,2)],
            columns=["turnover"],
            index=[self.fac_fwd],
        )
        return {"LS": turnover_LS, "top": turnover_top, "bottom": turnover_bottom}

    # 返回因子coverage图数据：月均coverage 或 周均coverage
    @timer
    def graph_coverage(self):
        df_cv = self.coverage().copy()
        df_cv = self.df_resample(df_cv, self.freq, "mean")
        def custom_round(x):
            if pd.notnull(x):
                return int(round(x))
            else:
                return x
        df_cv[self.fac_fwd] = df_cv[self.fac_fwd].apply(lambda x: custom_round(x))
        return df_cv

    # 返回因子IC表现图数据：月均ic及其rolling_avg 或 周均ic及其rolling_avg
    @timer
    def graph_ic(self):
        df_ic = self.df_ic.copy()
        df_ic = self.df_resample(df_ic, self.freq, "mean")
        df_ic[self.fac_fwd + "_avg"] = (
            df_ic[self.fac_fwd].rolling(window=self.window).mean()
        )
        return df_ic

    # 返回IC行业分布图数据：ic_mean
    @timer
    def graph_ic_industry(self):
        fac = self.fac.dropna(subset=[self.fac_name])
        fwd = self.fwd.dropna(subset=[self.fwd_ret_name])
        merged = pd.merge(fac, fwd, how="inner", on=["K", "D"])
        sd = merged["D"].min()
        ed = merged["D"].max()
        bin_list = list(range(1, 31))
        df_ic_ind = pd.DataFrame(
            {
                "bin": bin_list,
                "ind_code": IND_DICT.keys(),
                "ind_name": IND_DICT.values(),
                self.fac_fwd: "",
            },
            index=IND_DICT.keys(),
        )
        # 获取每个行业的成分股并计算ic
        for k, v in IND_DICT.items():
            ind = fqdatac.get_indconst(sd, ed, ind_code=k)
            fac_ind = pd.merge(ind, merged, how="inner", on=["K", "D"])

            def ic_group(group: pd.DataFrame) -> float:
                """
                计算T日的ic值
                """
                fac_rank = group[self.fac_name].rank(method="first")
                fwd_rank = group[self.fwd_ret_name].rank(method="first")
                rank_diff = fac_rank - fwd_rank
                n = len(group)
                if n <= 1:
                    res = np.nan
                else:
                    res = 1 - 6 * sum(rank_diff * rank_diff) / n / (n * n - 1)
                return res

            # 计算该行业的ic_mean
            ic_ind = fac_ind.groupby("D").apply(ic_group).mean()
            df_ic_ind.loc[(df_ic_ind["ind_code"] == k), self.fac_fwd] = ic_ind
        return df_ic_ind

    # 因子分十组的胜率
    @timer
    def graph_hit_rate_10bin(self):
        # 分组
        fac_bin = self.fac_bin.copy()
        # 合并因子值和收益率
        df_merge = pd.merge(fac_bin, self.fwd, on=["K", "D"], how="left")
        # 统计10组的r
        df_result = (
            df_merge.groupby("D")
            .agg(
                **{
                    f"bin{i}_0_cnt": (
                        self.fwd_ret_name,
                        lambda x, i=i: ((df_merge["bin"] == i) & (x > 0)).sum(),
                    )
                    for i in range(1, 11)
                },
                **{
                    f"bin{i}_cnt": ("bin", lambda x, i=i: (x == i).sum())
                    for i in range(1, 11)
                },
            )
            .reset_index()
        )
        # 计算每个 bin 的 hitrate
        bin_columns = [f"bin{i}" for i in range(1, 11)]
        df_result[bin_columns] = df_result[[f"bin{i}_0_cnt" for i in range(1, 11)]].div(
            df_result[[f"bin{i}_cnt" for i in range(1, 11)]].values
        )
        # 计算每个 bin 的均值
        df_hitrate = df_result[bin_columns].mean()
        df_hitrate = pd.DataFrame(
            {"bin": [i for i in range(1, 11)], self.fac_fwd: df_hitrate}
        )

        return df_hitrate

    # 单因子的月度换手率
    # 短于一个月返回日度，短于一年返回周度，长于一年返回月度
    @timer
    def graph_turnover_monthly(self):
        interval = self.get_date_interval(self.sd, self.ed)
        return self.turnover_different_freq(interval)

    @timer
    def graph_hitrate_monthly(self):
        interval = self.get_date_interval(self.sd, self.ed)
        return self.hitrate_different_freq(interval)

    def hitrate_different_freq(self, interval: str):
        hit_dict = self.hitrate()
        if interval == "day":
            roll = 5
        elif interval == "week":
            roll = 4
        else:
            roll = 6
        for k in list(hit_dict.keys()):
            df = hit_dict[k].copy()
            if interval != "day":
                df["D"] = pd.to_datetime(df["D"])  # 将'D'列转换为日期时间格式
                df.set_index("D", inplace=True)  # 设置'D'列为索引
                if interval == "week":
                    df = df.resample("W-FRI").mean()
                elif interval == "month":
                    df = df.resample("M").mean()
                df.reset_index(inplace=True)
                df.dropna(inplace=True)
                df["D"] = df["D"].dt.strftime("%Y%m%d")
            df[f"{self.fac_fwd}_avg"] = df[self.fac_fwd].rolling(roll).mean()
            hit_dict[k] = df
        return hit_dict

    def turnover_different_freq(self, interval: str):
        fac_bin = self.fac_bin.copy()
        bin_set = fac_bin.groupby(["D", "bin"])["K"].apply(set).reset_index()
        if interval != "day":
            bin_set["D"] = pd.to_datetime(bin_set["D"])  # 将'D'列转换为日期时间格式
            result = {}
            for bin_value, group in bin_set.groupby("bin"):
                group.set_index("D", inplace=True)  # 设置'D'列为索引
                if interval == "week":
                    resampled_last = group.resample("W-FRI").last()
                elif interval == "month":
                    resampled_last = group.resample("M").last()
                result[bin_value] = resampled_last
            fac_bin_resp = pd.DataFrame()
            for l in list(result.values()):
                fac_bin_resp = pd.concat([fac_bin_resp, l.reset_index()])
            fac_bin_resp["D"] = fac_bin_resp["D"].dt.strftime("%Y%m%d")
        else:
            fac_bin_resp = bin_set
        fac_bin_resp.dropna(inplace=True)
        # 获取日期列表
        dt_list = list(sorted(set(fac_bin_resp["D"])))
        result_LS, result_top, result_bottom = [], [], []
        prev_s1_set, prev_s10_set = set(), set()
        for dt in dt_list:
            # 如果为第一天，换手率为1
            if dt == dt_list[0]:
                result_LS.append([dt, 1])
                result_top.append([dt, 1])
                result_bottom.append([dt, 1])
                # 第一日的股票集合，给第二天用
                prev_s1_set = fac_bin_resp[
                    (fac_bin_resp["D"] == dt) & (fac_bin_resp["bin"] == 1)
                    ]["K"].iloc[0]
                prev_s10_set = fac_bin_resp[
                    (fac_bin_resp["D"] == dt) & (fac_bin_resp["bin"] == 10)
                    ]["K"].iloc[0]
                continue
            # dt_1 = dt_list[dt_list.index(dt)-1]
            # 当前日期的股票集合
            s1_set = fac_bin_resp[
                (fac_bin_resp["D"] == dt) & (fac_bin_resp["bin"] == 1)
                ]["K"].iloc[0]
            s10_set = fac_bin_resp[
                (fac_bin_resp["D"] == dt) & (fac_bin_resp["bin"] == 10)
                ]["K"].iloc[0]
            # T-1日的股票集合
            f1_set = prev_s1_set
            c1 = len(s1_set & f1_set)
            f10_set = prev_s10_set
            c10 = len(s10_set & f10_set)
            prev_s1_set, prev_s10_set = s1_set, s10_set
            if s1_set or s10_set:
                result_LS.append([dt, 1 - (c1 + c10) / (len(s1_set) + len(s10_set))])
                result_top.append([dt, 1 - c10 / len(s10_set)])
                result_bottom.append([dt, 1 - c1 / len(s1_set)])
            else:
                result_LS.append([dt, 0])
                result_top.append([dt, 0])
                result_bottom.append([dt, 0])
        df_LS = pd.DataFrame(result_LS, columns=["D", self.fac_fwd])
        df_top = pd.DataFrame(result_top, columns=["D", self.fac_fwd])
        df_bottom = pd.DataFrame(result_bottom, columns=["D", self.fac_fwd])
        if interval == "day":
            roll = 5
        elif interval == "week":
            roll = 4
        else:
            roll = 6
        df_LS[f"{self.fac_fwd}_avg"] = df_LS[self.fac_fwd].rolling(roll).mean()
        df_top[f"{self.fac_fwd}_avg"] = df_top[self.fac_fwd].rolling(roll).mean()
        df_bottom[f"{self.fac_fwd}_avg"] = df_bottom[self.fac_fwd].rolling(roll).mean()
        return {"LS": df_LS, "top": df_top, "bottom": df_bottom}

    # 计算持有期收益率
    def get_hpr(self, df_wealth, days, annualize):
        if days > len(df_wealth):
            temp_wealth = df_wealth[self.fac_fwd].copy()
        else:
            temp_wealth = df_wealth[self.fac_fwd].iloc[-days:]
        hpr = temp_wealth.iloc[-1] / temp_wealth.iloc[0] - 1
        if annualize:
            # n_year = len(temp_wealth) / 252
            # hpr = (1 + hpr) ** (1 / n_year) - 1
            daily_return = temp_wealth.pct_change()
            hpr = daily_return.mean() * (252**0.5)
            return hpr
        else:
            return hpr

    # 计算最大回撤
    def maxdrawdown(self, df_wealth):
        temp_wealth = df_wealth.sort_values(by="D", ascending=True).copy()
        temp_wealth.loc[:, "cummax"] = (
            temp_wealth[self.fac_fwd].to_frame().expanding().max()[self.fac_fwd]
        )
        temp_wealth.loc[:, "drawdown"] = (
                                                 temp_wealth["cummax"] - temp_wealth[self.fac_fwd]
                                         ) / temp_wealth["cummax"]
        max_drawdown = temp_wealth.loc[:, "drawdown"].max()
        return max_drawdown

    # 计算综合performance
    def ret_performance(self, df_wealth, df_return):
        # return_1m: 回测期最后21个交易日的return（%）
        return_1m = self.get_hpr(df_wealth=df_wealth, days=21, annualize=False)
        # return_3m: 回测期最后63个交易日的return（%）
        return_3m = self.get_hpr(df_wealth=df_wealth, days=53, annualize=False)
        # cagr_1y: 回测期最后252个交易日的return（%）
        cagr_1y = self.get_hpr(df_wealth=df_wealth, days=252, annualize=False)
        # cagr_3y: 回测期最后回测期最后756个交易日的年化return（%）
        cagr_3y = self.get_hpr(df_wealth=df_wealth, days=756, annualize=True)
        # cagr_5y: 回测期最后回测期最后1260个交易日的年化return（%）
        cagr_5y = self.get_hpr(df_wealth=df_wealth, days=1260, annualize=True)
        # cagr: 回测期的return（%)
        cagr = self.get_hpr(df_wealth=df_wealth, days=len(df_wealth), annualize=True)
        # volatility: 回测期因子组合收益的波动率（%）
        volatility = df_return[self.fac_fwd].std() * np.sqrt(252)
        # sharpe_ratio:回测期因子组合的夏普比率, 暂时用rf = 0.02
        sharp_ratio = (cagr - 0.02) / volatility
        # maximum drawdown:回测期因子组合的最大回撤
        max_drawdown = self.maxdrawdown(df_wealth)
        performance = pd.DataFrame(
            {
                "return_1m": round(return_1m * 100, 2),
                "return_3m": round(return_3m * 100, 2),
                "cagr_1y": round(cagr_1y * 100, 2),
                "cagr_3y": round(cagr_3y * 100, 2),
                "cagr_5y": round(cagr_5y * 100, 2),
                "cagr": round(cagr * 100, 2),
                "volatility": round(volatility * 100, 2),
                "sharp_ratio": round(sharp_ratio, 2),
                "max_drawdown": round(max_drawdown * 100, 2),
            },
            index=[self.fac_fwd],
        )
        return performance

    # 返回LS、top、bottom三组summary_ret的1月，3月，1年，3年，5年，收益率，波动率，sharp ratio
    @timer
    def summary_ret(self):
        # LS 组合的performance
        ls_performance = self.ret_performance(self.df_wealth, self.LSreturn)
        # top 组合的performance
        top_performance = self.ret_performance(self.top_wealth, self.rk)
        # bottom 组合的performance
        bottom_performance = self.ret_performance(self.bottom_wealth, self.rs)
        # 分10组的perfromance：
        group_perfromance = []
        for bin in set(self.wealth_by_group["bin"].to_list()):
            sub_wealth = self.wealth_by_group[self.wealth_by_group["bin"] == bin][
                ["D", self.fac_fwd]
            ]
            sub_ret = self.ret_all[self.ret_all["bin"] == bin][["D", self.fac_fwd]]
            sub_performance = self.ret_performance(sub_wealth, sub_ret)
            sub_performance.insert(loc=0, column="bin", value=[bin])
            sub_performance = sub_performance.reset_index(drop=True)
            sub_performance.insert(loc=0, column="fac_name", value=[self.fac_fwd])
            group_perfromance.append(sub_performance)
        group_perfromance = pd.concat(group_perfromance, axis=0)
        return {
            "LS": ls_performance,
            "top": top_performance,
            "bottom": bottom_performance,
        }, group_perfromance

    # 计算分年度cagr
    @timer
    def cagr_by_year(self, df_wealth):
        temp_wealth = df_wealth.copy().sort_values(by="D", ascending=True)
        temp_wealth["year"] = temp_wealth["D"].str.slice(0, 4)
        col_name = []
        ret_year = []
        for year in temp_wealth["year"].unique():
            sub_wealth = temp_wealth[temp_wealth["year"] == year].copy()
            ret_i = (
                    sub_wealth[self.fac_fwd].iloc[-1] / sub_wealth[self.fac_fwd].iloc[0] - 1
            )
            col_name.append(year)
            ret_year.append(round(ret_i * 100, 2))
        col_name = ["cagr_" + i for i in col_name]
        data = dict(zip(col_name, ret_year))
        df_cagr_by_year = pd.DataFrame(data, index=[self.fac_fwd])
        return df_cagr_by_year

    # 返回LS、top、bottom三组cagt_by_year
    @timer
    def summary_cagr_year_t(self):
        ls_cagr_by_year = self.cagr_by_year(self.df_wealth)
        top_cagr_by_year = self.cagr_by_year(self.top_wealth)
        bottom_cagr_by_year = self.cagr_by_year(self.bottom_wealth)
        return {
            "LS": ls_cagr_by_year,
            "top": top_cagr_by_year,
            "bottom": bottom_cagr_by_year,
        }

    # 计算平均serial correlation
    @timer
    def summary_sc(self):
        avg_sc = self.serial_correlation[self.fac_fwd].mean()
        summary_sc = pd.DataFrame(
            {"serial_corr": round(avg_sc * 100, 2)}, index=[self.fac_fwd]
        )
        return summary_sc

    # 获得LS，top，bottom的summary report
    @timer
    def get_summary(self):
        # coverage_max, coverage_min
        df_coverage = self.summary_coverage()

        # return_1m,return_3m,cagr_1y,cagr_3y,cagr_5y,cagr,volatility,sharpe_ratio,max_drawdown
        dict_ret, self.performace_by_group = self.summary_ret()

        # hit_rate
        dict_hit_rate = self.summary_hit_rate()

        # rank_ic_mean, rank_ic_std, ic_ir
        df_ic_ir = self.summary_ic_ir()

        # serial correlation
        df_sc = self.summary_sc()

        # turnover
        dict_turnover = self.summary_turnover()

        # cagr_year_t
        dict_cagr_year = self.summary_cagr_year_t()

        # rank_ic_mean_year: 列数量与回测区间包含的年度数量有关
        df_ic_year = self.summary_ic_year()

        summary = {}
        for key in ["LS", "top", "bottom"]:
            all_df = [
                df_coverage,
                dict_ret[key],
                dict_hit_rate[key],
                df_ic_ir,
                df_sc,
                dict_turnover[key],
                dict_cagr_year[key],
                df_ic_year,
            ]
            sub_summary = pd.concat(all_df, axis=1)
            sub_summary.insert(loc=0, column="backtest_ed", value=[self.ed])
            sub_summary.insert(loc=0, column="backtest_sd", value=[self.sd])
            sub_summary = sub_summary.reset_index(drop=True)
            sub_summary.insert(loc=0, column="fac_name", value=[self.fac_fwd])
            summary[key] = sub_summary
        return summary



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: api_fqdatac.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/31 15:26:25
"""

import datetime
import pandas as pd
from dateutil.parser import parse
from rqalpha.api import export_as_api
from rqalpha.utils.arg_checker import apply_rules, verify_that
from ..core.exception import FqDatacException

try:
    import fqdatac
except ImportError:
    class DummyFqDatac:
        __name__ = "fqdatac"

        def __getattr__(self, item):
            return self

        def __call__(self, *args, **kwargs):
            raise FqDatacException('fqdatac is not available, extension apis will not function properly')
    fqdatac = DummyFqDatac()


def to_date(date):
    if isinstance(date, datetime.datetime):
        return date.date()
    if isinstance(date, datetime.date):
        return date

    if isinstance(date, str):
        return parse(date).date()

    raise FqDatacException('unknown date value: {}'.format(date))


def to_date_str(date):
    if isinstance(date, datetime.datetime):
        return date.strftime('%Y%m%d')
    elif isinstance(date, datetime.date):
        return date.strftime('%Y%m%d')
    elif isinstance(date, str):
        return date.replace("-", "")
    return date


@export_as_api
@apply_rules(verify_that('start_date').is_valid_date())
@apply_rules(verify_that('end_date').is_valid_date())
def get_daysrange(start_date, end_date) -> list:
    """
    从fqdatac取数据
    :param start_date:
    :param end_date:
    :return:
    """
    start_date = to_date_str(start_date)
    end_date = to_date_str(end_date)
    #df_days = pd.read_csv('/Users/wangjiangfeng/.faithquant/fakedata/all_days.csv', header=0)
    #trade_days = df_days['day'].tolist()
    #return [str(day) for day in trade_days if start_date <= str(day) <= end_date]
    return fqdatac.get_daysrange(start_date, end_date, days_type='trading', output_format='str')


@export_as_api
@apply_rules(verify_that('start_date').is_valid_date())
@apply_rules(verify_that('end_date').is_valid_date())
def get_universe(start_date, end_date, univ_name) -> pd.DataFrame:
    """
    从fqdatac取数据
    :param start_date: yyyyMMdd
    :param end_date: yyyyMMdd
    :param univ_name: universe name
    :return: pd.DataFrame
    """
    return fqdatac.get_universe(start_date, end_date, univ_name)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: strategy_day_bar.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/7 15:33:26
"""

from rqalpha.apis import *


def init(context):
    context.fired = False
    logger.info("Fatbulls组合回测框架初始化")


def before_trading(context):
    pass


def handle_bar(context, bar_dict):
    """你选择的证券的数据更新将会触发此段逻辑，例如日或分钟历史数据切片或者是实时数据切片更新"""
    bar_date = context.now
    day = bar_date.strftime('%Y%m%d')
    logger.info("handle_bar开始处理")
    if hasattr(context.config.faithquant.portfolio.stock_positions, day):
        df_stocks = getattr(context.config.faithquant.portfolio.stock_positions, day)
        if df_stocks is None or df_stocks.empty:
            logger.info("组合数据为空，执行清仓操作")
            logger.info("开始执行清仓操作")
            positions = get_positions()
            for position in positions:
                order_book_id = position.order_book_id
                quantity = position.quantity
                order_shares(order_book_id, quantity * -1)
        else:
            target_portfolio = {k: w for k, w in zip(df_stocks['K'], df_stocks['W'])}
            price_or_styles = {k: p for k, p in zip(df_stocks['K'], df_stocks['vwap'])}
            # 调用目标函数按目标下单
            if not context.fired:
                order_target_portfolio(target_portfolio, price_or_styles)
                context.fired = True
    else:
        logger.info("持仓数据不存在，日期：{}", day)
    logger.info("交易完毕")


def after_trading(context):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: logger.py
Author: wuyiyang(wuyiyang@hcyjs.com)
Date: 2023/8/14 16:37:56
"""

import os
import logging
import datetime as dt
from pathlib import Path


def init_logging(app_name, app_dir=None, console_level=logging.INFO):
    if app_dir is None:
        app_dir = Path(__file__).absolute().parent.parent.parent

    logdir = os.path.join(app_dir, 'log', app_name)
    os.makedirs(logdir, exist_ok=True)
    today_date: str = dt.datetime.now().strftime("%Y%m%d")
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    for handler in list(logger.handlers):
        logger.removeHandler(handler)

    fmt = logging.Formatter("%(asctime)s [%(name)s] [%(levelname)s]: %(message)s")

    fn: str = os.path.join(logdir, f"{app_name}_{today_date}.log")
    file_handler = logging.FileHandler(fn, mode="a", encoding="utf8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(fmt)
    logger.addHandler(file_handler)

    warn_fn: str = os.path.join(logdir, f"{app_name}_{today_date}_warn.log")
    warnfile_handler = logging.FileHandler(warn_fn, mode="a", encoding="utf8")
    warnfile_handler.setLevel(logging.WARNING)
    warnfile_handler.setFormatter(fmt)
    logger.addHandler(warnfile_handler)

    console = logging.StreamHandler()
    console.setLevel(console_level)
    console.setFormatter(fmt)
    logger.addHandler(console)


def check_exist_and_makedir(path,file_name):
    files = os.listdir(path)
    current_file = [x for x in files ]
    folder_path = os.path.join(path, file_name)
    if file_name in current_file:
        pass
    else:
        os.mkdir(folder_path)
    return folder_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


from .logger import init_logging
import logging
import fire
from . import fac_class as fac
from pathlib import Path
import os


def backtest_main(
        backtest_name: str,
        sd: str,
        ed: str,
        task_num: str,
        univ_name: str = "cnall",
        fdays: int = 1,
        rettype: str = "vwap30 to vwap30",
        directory_list: list = [],
):
    # 实例化
    loader = fac.FactorBTLoader(
        sd=sd,
        ed=ed,
        backtest_name=backtest_name,
        univ_name=univ_name,
        fdays=fdays,
        rettype=rettype,
        directory_list=directory_list,
    )

    # 数据加载与预处理
    logging.info("start factor loading")
    for d in directory_list:
        loader.load(
            directory=d,
        )
        logging.info(f"factor {d.split('/')[-1]} successfully loaded")
    loader.onlyuniv()

    logging.info("start fwdret loading")
    loader.fwdret_load()

    # 整个回测过程
    logging.info("start factor evaluation")
    output_path = "/data/quant/report/factor_backtest"
    evaluator = fac.FactorEvaluator(
        sd=loader.sd,
        ed=loader.ed,
        backtest_name=loader.backtest_name,
        univ_name=loader.univ_name,
        fdays=loader.fdays,
        rettype=loader.rettype,
        directory_list=loader.directory_list,
        raw=loader.container,
        fwdret1=loader.fwdret1,
        fwdret = loader.fwdret,
        output_path=output_path,
        task_num=task_num,
    )
    evaluator.initialize()
    evaluator.nextfactor()
    while evaluator.whole_processing:
        # evaluator.nextfile()
        # while evaluator.fac_processing:
        #     evaluator.evaluate()
        #     evaluator.save_lastfile()
        #     evaluator.nextfile()
        logging.info(f"factor evaluation for factor{evaluator.curfac}")
        evaluator.evaluate()
        evaluator.fac_report()
        evaluator.nextfactor()
        logging.info("--next factor")
    evaluator.output()


def run(
        backtest_name: str,
        sd: str,
        ed: str,
        task_num: str = "b",
        univ_name: str = "cnall",
        fdays: int = 5,
        rettype: str = "vwap30",
        directory_list: list = [],
):
    fac.FQDataLoader()

    output_path = "/data/quant/report/factor_backtest"
    # 建立sql连接
    # sql_path = os.path.join(
    #     Path(__file__).absolute().parent.parent.parent.parent,
    #     "sqlconn",
    #     "connection.txt",
    # )
    # conndict = SQLConnection.read_connfile(path=sql_path)
    # conn = SQLConnection.load_connection(d=conndict)
    # conn.mysql_connection()

    backtest_main(
        backtest_name=backtest_name,
        sd=sd,
        ed=ed,
        task_num=task_num,
        univ_name=univ_name,
        fdays=fdays,
        rettype=rettype,
        directory_list=directory_list,
    )


if __name__ == "__main__":
    DEBUG = True
    init_logging(__file__.split("/")[-1].split(".")[0])
    logging.info("---start factor evaluation---")
    if DEBUG:
        run(
            backtest_name="test_backtest1",
            sd="20170101",
            ed="20201231",
            directory_list=[
                #"/data/quant/factors/DTech/DTech_volume",
                "/data/quant/factors/DTech/DTech_amt",
                "/data/quant/factors/DTech/DTech_high",
                # "/data/quant/factors/DTech/DTech_ret",
                # "/data/quant/factors/DTech/DTech_lret",
                # "/data/quant/factors/DTech/DTech_opn",
                # "/data/quant/factors/DTech/DTech_vwap",
            ],
        )

        ...
    else:
        fire.Fire(run)

    logging.warning(f"All done")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: const.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/12 11:13:31
"""

from rqalpha.const import CustomEnum


# noinspection PyPep8Naming
class BACKTEST_TYPE(CustomEnum):

    FACTOR_BACKTEST = 'FACTORS'

    PORTFOLIO_BACKTEST = 'PORTFOLIOS'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################




"""
File: __init__.py.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 16:36:59
"""

import os
import re

FQDATAC_DEFAULT_ADDRESS = "fqdatac://ak:sk@219.143.244.230:8390"


def init_fqdatac_env(uri):
    if uri is None:
        return

    if '@' not in uri:
        uri = "tcp://{}@{}".format(uri, FQDATAC_DEFAULT_ADDRESS)

    if not re.match(r"\w*://.+:.+@.+:\d+", uri):
        raise ValueError('invalid fqdatac uri. use user:password or tcp://user:password@ip:port')

    os.environ['FQDATAC_CONF'] = uri



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: __init__.py.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 16:09:11
"""
import time
import logging
import os
from functools import wraps

__all__ = [
    '__version__', 'run_backtest', 'time_cost'
]


def time_cost(func):

    @wraps(func)
    def wrap(*args, **kwargs):
        begin_time = time.perf_counter()
        result = func(*args, **kwargs)
        start_time = time.perf_counter()
        tips = '函数: %r 耗时: %2.4f 秒' % (func.__name__, start_time - begin_time)
        logging.info(tips)
        print(tips)
        return result
    return wrap


@time_cost
def run_backtest(**kwargs):
    """
    传入约定函数和因子配置运行回测。约定函数详见 API 手册约定函数部分，可用的配置项详见参数配置部分。

    :Keyword Arguments:
        * **config** (dict) -- 策略配置字典

    :return: dict

    """
    from fatbulls.utils.const import BACKTEST_TYPE
    from fatbulls import main
    backtest_type = kwargs.get('backtest_type', None)
    if backtest_type is None:
        raise ValueError("missing parameter: backtest_type")
    backtest_config = kwargs.get('backtest_config', {})
    if backtest_type == BACKTEST_TYPE.FACTOR_BACKTEST:
        """
        因子回测方法调用
        """
        return main.run_factor_backtest_se(backtest_config)
    elif backtest_type == BACKTEST_TYPE.PORTFOLIO_BACKTEST:
        """ 组合回测的逻辑 """
        from rqalpha.utils.functools import clear_all_cached_functions
        from rqalpha.utils.config import load_yaml
        from rqalpha.utils.dict_func import deep_update
        from fatbulls.utils.request import PortfolioRequest

        fatbulls_config_path = kwargs.get('config_path', None)
        if fatbulls_config_path is None:
            fatbulls_config_path = os.path.join(os.path.dirname(__file__), 'resource/fatbulls_config.yml')
        config = load_yaml(fatbulls_config_path)
        pr = PortfolioRequest(**backtest_config)
        backtest_config = pr.get_config()
        deep_update(backtest_config, config)
        clear_all_cached_functions()
        return main.run_portfolio_backtest(config=config)
    else:
        raise ValueError("invalid parameter: backtest_type, It could be FACTORS or PORTFOLIOS")


__version__ = '0.1.1-dev'





#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################

import os

"""
File: request.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/5 16:45:25
"""

RETURN_TYPE = ['vwap30', 'vwap15', 'o2o', 'c2c']

class PortfolioRequest(object):

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_config(self):
        config = dict()
        config['base'] = self.get_base_config()
        config['mod'] = {}
        config['mod']['sys_analyser'] = self.get_analyser_config()
        config['mod']['sys_simulation'] = self.get_simulation_config()
        config['mod']['sys_transaction_cost'] = self.get_transaction_cos_config()
        config['faithquant'] = {}
        config['faithquant']['portfolio'] = self.get_portfolio_config()
        return config

    def get_start_date(self):
        if hasattr(self, 'sd'):
            sd = getattr(self, 'sd')
        else:
            raise ValueError('paramter sd not found, It could be a start date, eg: 20230331')
        # return datetime.strptime(sd, "%Y%m%d").strftime("%Y-%m-%d")
        return sd

    def get_end_date(self):
        if hasattr(self, 'ed'):
            ed = getattr(self, 'ed')
        else:
            raise ValueError('paramter ed not found, It could be a end date, eg: 20230331')
        # return datetime.strptime(ed, "%Y%m%d").strftime("%Y-%m-%d")
        return ed

    def get_stock_accounts(self):
        if hasattr(self, 'stock_accounts'):
            stock_accounts = getattr(self, 'stock_accounts')
        else:
            raise ValueError('paramter stock_accounts not found, It could be a integer, eg: 10000000000')
        return stock_accounts

    def get_benchmark(self):
        if hasattr(self, 'benchmark'):
            benchmark = getattr(self, 'benchmark')
        else:
            raise ValueError('paramter benchmark not found, It could be a instrument code, eg: 000300.XSHG')
        return benchmark

    def get_matching_type(self):
        if hasattr(self, 'matching_type'):
            matching_type = getattr(self, 'matching_type')
        else:
            raise ValueError('paramter matching_type not found, It could be a enum value, eg: vwap, current_bar')
        return matching_type

    def get_slippage(self):
        if hasattr(self, 'slippage'):
            slippage = getattr(self, 'slippage')
        else:
            raise ValueError('paramter slippage not found, It could be a integer, eg: 1')
        return slippage

    def get_stock_commission_multiplier(self):
        if hasattr(self, 'stock_commission_multiplier'):
            stock_commission_multiplier = getattr(self, 'stock_commission_multiplier')
        else:
            raise ValueError('paramter stock_commission_multiplier not found, It could be a integer, eg: 1')
        return stock_commission_multiplier

    def get_tax_multiplier(self):
        if hasattr(self, 'tax_multiplier'):
            tax_multiplier = getattr(self, 'tax_multiplier')
        else:
            raise ValueError('paramter tax_multiplier not found, It could be a integer, eg: 1')
        return tax_multiplier

    def get_management_fee(self):
        if hasattr(self, 'management_fee'):
            management_fee = getattr(self, 'management_fee')
        else:
            raise ValueError('paramter management_fee not found, It could be a integer, eg: 1')
        management_fee = [("STOCK", management_fee), ("FUTURE", 0.0000)]
        return management_fee

    def get_base_config(self):
        base_config = dict()
        base_config['start_date'] = self.get_start_date()
        base_config['end_date'] = self.get_end_date()
        base_config['accounts'] = {}
        base_config['accounts']['stock'] = self.get_stock_accounts()
        base_config['run_type'] = self.get_run_type()
        return base_config

    def get_analyser_config(self):
        """
        'sys_analyser': {
                    'benchmark': '',
                    'output_file': '',
                    'report_save_path': '',
                    'plot_save_file': '',
                    'plot_config': {
                        'open_close_points': True,
                        'weekly_indicators': True
                    },
                },
                 "record": True,
            "output_file": "/Users/wangjiangfeng/Downloads/demo.pkl",
            "report_save_path": "/Users/wangjiangfeng/Downloads",
            "enabled": True,
            "plot": True,
            "plot_save_file": "/Users/wangjiangfeng/Downloads",
        """
        analyser_config = dict()
        analyser_config['record'] = True
        analyser_config['benchmark'] = self.get_benchmark()
        analyser_config['output_file'] = ''
        analyser_config['report_save_path'] = ''
        analyser_config['plot_save_file'] = ''
        analyser_config['plot_config'] = {
            'open_close_points': True,
            'weekly_indicators': True
        }
        return analyser_config

    def get_simulation_config(self):
        """
        'sys_simulation': {
                    'matching_type': '',
                    'slippage': '',
                    'anagement_fee': []
                },
        """
        simulation_config = dict()
        # @TODO 暂时用固定值
        simulation_config['matching_type'] = 'current_bar'
        simulation_config['slippage'] = self.get_slippage()
        simulation_config['management_fee'] = self.get_management_fee()
        return simulation_config

    def get_transaction_cos_config(self):
        """
        'sys_transaction_cost': {
                    'commission_multiplier': '',
                    'stock_commission_multiplier': '',
                    'tax_multiplier': '',
                }
        """
        transaction_cost = dict()
        transaction_cost['stock_commission_multiplier'] = self.get_stock_commission_multiplier()
        transaction_cost['tax_multiplier'] = self.get_tax_multiplier()
        return transaction_cost

    def get_portfolio_config(self):
        portfolio_config = dict()
        portfolio_config['portfolio_dirpath'] = self.get_portfolio_dirpath()
        portfolio_config['base_report_path'] = self.get_base_report_path()
        portfolio_config['task_num'] = self.get_task_num()
        portfolio_config['backtest_name'] = self.get_backtest_name()
        portfolio_config['return_type'] = self.get_return_type()
        return portfolio_config

    def get_portfolio_dirpath(self):
        if hasattr(self, 'portfolio_dirpath'):
            portfolio_dirpath = getattr(self, 'portfolio_dirpath')
        else:
            raise ValueError('paramter portfolio_dirpath not found, It could be a filepath,'
                             ' eg: /data/quant/portfolios/wuyiyang/baseline_1001/portfolios')
        if not os.path.exists(portfolio_dirpath):
            raise NotADirectoryError('portfolio_dirpath not exists')
        return portfolio_dirpath

    def get_base_report_path(self):
        base_report_path = ''
        if hasattr(self, 'base_report_path'):
            base_report_path = getattr(self, 'base_report_path', None)
            if (base_report_path is not None) and (not os.path.exists(base_report_path)):
                raise NotADirectoryError('base_report_path not exists')
        return base_report_path

    def get_backtest_name(self):
        if hasattr(self, 'backtest_name'):
            backtest_name = getattr(self, 'backtest_name', None)
        else:
            raise ValueError('paramter backtest_name not found, It could be a string, eg: 房地产回测')
        return backtest_name

    def get_task_num(self):
        if hasattr(self, 'task_num'):
            task_num = getattr(self, 'task_num', None)
        else:
            raise ValueError('paramter task_num not found, It could be a unique string, eg: 1123333')
        return task_num

    def get_run_type(self):
        if hasattr(self, 'run_type'):
            run_type = getattr(self, 'run_type')
        else:
            run_type = 'b'
        return run_type

    def get_return_type(self):
        if not hasattr(self, 'return_type'):
            raise ValueError('paramter run_type not found, It could be a enum string, eg: o2o, c2c, vwap15...')
        else:
            return_type = getattr(self, 'return_type')
        if return_type not in RETURN_TYPE:
            raise ValueError('invalid run_type, It could be a enum string, eg: o2o, c2c, vwap15...')
        return return_type



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: portfolio.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/6 15:08:53
"""

import os
import pandas as pd
from rqalpha.utils.repr import property_repr
from fatbulls.core.exception import FactorBackTestException
from fatbulls.portfolio.constant import PORTFOLIO_COLUMNS, PORTFOLIO_MAX_WEIGHT, PORTFOLIO_WEIGHT_FIELD
from rqalpha.utils.functools import lru_cache


class Portfolios(object):

    __repr__ = property_repr

    def __init__(self, portfolio_name, portfolio_dirpath):
        self._portfolio_name = portfolio_name
        self._portfolio_dirpath = portfolio_dirpath

    def check_portfolio(self, trade_days):
        days = self.get_portfolio_days()
        if set(trade_days).issubset(days):
            return True
        else:
            missing_days = [str(day) for day in trade_days if str(day) not in days]
            raise FactorBackTestException("missing trade day csv file in portfolio path, days: {}"
                                          .format(",".join(missing_days)))

    @property
    def portfolio_name(self):
        return self._portfolio_name

    @property
    def portfolio_dirpath(self):
        return self._portfolio_dirpath

    @lru_cache(128)
    def get_portfolio_days(self):
        return [file.split(".")[0] for file in os.listdir(self._portfolio_dirpath)]

    @lru_cache(128)
    def get_portfolio_data(self, start_date=None, end_date=None) -> dict:
        if not os.path.exists(self._portfolio_dirpath):
            raise FactorBackTestException("portfolio dirpath not exists")
        portfolio_data = {}
        portfolio_days = self.get_portfolio_days()
        max_day = max(portfolio_days)
        min_day = min(portfolio_days)
        start_date = min_day if start_date is None else start_date
        end_date = max_day if end_date is None else end_date
        for trade_day in portfolio_days:
            if trade_day < start_date or trade_day > end_date:
                continue
            file = "{}.csv".format(trade_day)
            filename = os.path.join(self._portfolio_dirpath, file)
            df_portfolio_day = pd.read_csv(filename, header=0)
            if not df_portfolio_day.empty:
                sum_weight = df_portfolio_day[PORTFOLIO_WEIGHT_FIELD].sum()
                if sum_weight > PORTFOLIO_MAX_WEIGHT:
                    """总权重压缩到0到0.95区间"""
                    df_portfolio_day[PORTFOLIO_WEIGHT_FIELD] = df_portfolio_day[PORTFOLIO_WEIGHT_FIELD] / sum_weight
                    df_portfolio_day[PORTFOLIO_WEIGHT_FIELD] = (df_portfolio_day[PORTFOLIO_WEIGHT_FIELD]
                                                                * (PORTFOLIO_MAX_WEIGHT - 0.0) + 0.0)
                df_portfolio_day['K'] = df_portfolio_day['K'].apply(lambda x: str(x)
                                                                    .replace(".SH", ".XSHG")
                                                                    .replace(".SZ", ".XSHE"))
                # 保留4为小数
                df_portfolio_day['W'] = df_portfolio_day['W'].round(4)
                df_portfolio_day['vwap'] = df_portfolio_day['vwap'].round(4)
                portfolio_data[trade_day] = df_portfolio_day[['K', 'D', 'T', 'W', 'vwap']]
            else:
                portfolio_data[trade_day] = {}
        return portfolio_data


# if __name__ == "__main__":
#     p = Portfolios('baseline001', '/Users/wangjiangfeng/.faithquant/portfolios')
#     p.get_portfolio_data()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: run.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 17:22:48
"""


import os

import click
from rqalpha.utils.click_helper import Date
from rqalpha.utils.config import parse_config

from .entry import cli


@cli.command(help="Run a strategy")
@click.help_option('-h', '--help')
# -- Base Configuration
@click.option('-d', '--data-bundle-path', 'base__data_bundle_path', type=click.Path(exists=True))
@click.option('-f', '--strategy-file', 'base__strategy_file', type=click.Path(exists=True))
@click.option('-s', '--start-date', 'base__start_date', type=Date())
@click.option('-e', '--end-date', 'base__end_date', type=Date())
@click.option('-l', '--log-level', 'extra__log_level', type=click.Choice(['verbose', 'debug', 'info', 'error', 'none']))
def run(**kwargs):
    config_path = kwargs.get('config_path', None)
    if config_path is not None:
        config_path = os.path.abspath(config_path)
        kwargs.pop('config_path')
    if not kwargs.get('base__securities', None):
        kwargs.pop('base__securities', None)

    from fatbulls import main
    source_code = kwargs.get("base__source_code")
    cfg = parse_config(kwargs, config_path=config_path, click_type=True, source_code=source_code)
    source_code = cfg.base.source_code
    results = main.run(cfg, source_code=source_code)

    if results is None:
        return 1


def inject_run_param(param: click.Parameter):
    cli.commands["run"].params.append(param)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: exception.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/31 15:36:50
"""


class FactorBackTestException(Exception):
    """
    因子回测通用异常类
    """
    def __init__(self, msg):
        self.msg = msg


class FqDatacException(Exception):
    """
    SDK调用异常
    """
    def __init__(self, msg):
        self.msg = msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: context.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 16:33:27
"""
import logging
from rqalpha.core.global_var import GlobalVars
from fatbulls.model.config import FactorBackTestConfig


class FatbullsFactorContext(object):
    _fbctx = None  # type: FatbullsFactorContext

    def __init__(self, config):
        FatbullsFactorContext._fbctx = self
        self.config = config
        self.global_vars = GlobalVars()
        self.backtest_config = None

    @classmethod
    def get_instance(cls):
        """
        返回已经创建的 Context 对象
        """
        if FatbullsFactorContext._fbctx is None:
            raise RuntimeError(
                "FatbullsFactorContext has not been created."
                " Please Use `FatbullsFactorContext.get_instance()` after FatBulls init")
        return FatbullsFactorContext._fbctx

    def set_config(self, config):
        self.config = config

    def get_config(self):
        return self.config

    def get_factor_config(self) -> FactorBackTestConfig:
        try:
            if self.backtest_config is not None:
                return self.backtest_config
            self.backtest_config = FactorBackTestConfig(
                sd=self.config.factor.sd,
                ed=self.config.factor.ed,
                task_num=self.config.factor.task_num,
                backtest_name=self.config.factor.backtest_name,
                univ_name=self.config.factor.univ_name,
                fdays=self.config.factor.fdays,
                rettype=self.config.factor.rettype,
                factor_dirs=self.config.factor.factor_dirs
            )
        except Exception as e:
            raise EnvironmentError('valid factor backtest config!, msg: {}'.format(str(e)))
        return self.backtest_config

    def get_base_logpath(self):
        try:
            return self.config.factor.base_log_path
        except Exception as e:
            logging.warning("factor.base_log_path not exists")
            return None

    def get_base_report_dirpath(self):
        try:
            return self.config.factor.base_report_dirpath
        except Exception as e:
            raise FileNotFoundError("factor.base_report_dirpath not exists")


class FatbullsPortfolioContext(object):
    _fbctx = None  # type: FatbullsPortfolioContext

    def __init__(self, config):
        FatbullsPortfolioContext._fbctx = self
        self.config = config
        self.global_vars = GlobalVars()
        self.backtest_config = None

    @classmethod
    def get_instance(cls):
        """
        返回已经创建的 Context 对象
        """
        if FatbullsPortfolioContext._fbctx is None:
            raise RuntimeError(
                "FatbullsPortfolioContext has not been created."
                " Please Use `FatbullsPortfolioContext.get_instance()` after FatBulls init")
        return FatbullsPortfolioContext._fbctx




#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: strategy_day_bar.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/7 15:33:26
"""

from rqalpha.apis import *


# 在这个方法中编写任何的初始化逻辑。context对象将会在你的算法策略的任何方法之间做传递。
def init(context):
    logger.info("Fatbulls组合回测框架初始化")
    # 是否已发送了order
    context.fired = False


def before_trading(context):
    pass


def handle_bar(context, bar_dict):
    """你选择的证券的数据更新将会触发此段逻辑，例如日或分钟历史数据切片或者是实时数据切片更新"""
    #
    if not context.fired:
        # order_percent并且传入1代表买入该股票并且使其占有投资组合的100%
        order_percent(context.s1, 1)
        context.fired = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: main.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 17:30:51
"""

import os
import logging
import pandas as pd

from deprecated import deprecated
from fatbulls.utils import init_fqdatac_env
from fatbulls.utils.logger import init_logging
from fatbulls.factor.se import fac_backtest
from fatbulls.core.context import FatbullsFactorContext
from fatbulls.portfolio.backtest import PortfolioBacktestEngine
from fatbulls.core.exception import FactorBackTestException


def set_pandas_options():
    pd.set_option("display.max_columns", 100)
    pd.set_option("display.width", 1000)
    pd.set_option("display.max_colwidth", 100)
    pd.set_option("display.float_format", lambda x: "%.4f" % x)


def init_fqdatac(fqdatac_uri):
    if fqdatac_uri in ["disabled", "DISABLED"]:
        return
    try:
        import fqdatac
    except ImportError:
        return
    if isinstance(fqdatac.client.get_client(), fqdatac.client.DummyClient):
        init_fqdatac_env(fqdatac_uri)
        try:
            fqdatac.init()
        except Exception as e:
            logging.error('fqdatac init failed, some apis will not function properly: {}'.format(str(e)))


def run_factor_backtest_se(config, app_name='fatbulls_factor_backtest', set_pandas=True):
    try:
        """
        config = {
            backtest_name: str,
            sd: str,
            ed: str,
            task_num: str = "b",
            univ_name: str = "cnall",
            fdays: int = 5,
            rettype: str = "vwap30",
            directory_list: list = [],
        }
        """
        logging.info("app_name: {}", app_name)
        logging.info("set_pandas: {}", set_pandas)
        # 不做任何处理直接调用
        fac_backtest.run(**config)
    except Exception as e:
        raise e


@deprecated
def run_factor_backtest(config, app_name='fatbulls_factor_backtest', set_pandas=True):
    try:
        if set_pandas:
            """pandas 设置"""
            set_pandas_options()
        init_fqdatac(config['base']['fqdatac_uri'])
        if config.factor.base_log_path:
            app_dir = config.factor.base_log_path
        else:
            app_dir = None
        init_logging(app_name, app_dir)
        fat_ctx = FatbullsFactorContext(config)
        factor_conf = fat_ctx.get_factor_config()
        fat_ctx.get_base_report_dirpath()
        logging.info(factor_conf)
        logging.info("---start factor evaluation---")
        logging.info("mock execute")
        logging.info("All done")
    except Exception as e:
        raise e


def run_portfolio_backtest(config, set_pandas=True):
    try:
        if set_pandas:
            """pandas 设置"""
            set_pandas_options()
        init_fqdatac(config['base']['fqdatac_uri'])
        run_type = config['base']['run_type']
        frequency = config['base']['frequency']
        if run_type == 'b' and frequency == '1d':
            # 目前只支持日度回测
            base_path = os.path.dirname(os.path.abspath(__file__))
            strategy_file_path = os.path.join(base_path, "portfolio/strategy_day_bar.py")
            engine = PortfolioBacktestEngine(backtest_config=config, strategy_file_path=strategy_file_path)
            engine.run()
        else:
            raise FactorBackTestException("only support run type of stock history backtest and daily frequency")
        logging.info("All done")
    except Exception as e:
        raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: backtest.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/9/6 14:00:37
"""

import os
from fatbulls.core.engine import AbstractBackTestEngine
from fatbulls.core.exception import FactorBackTestException
from fatbulls.model.portfolio import Portfolios
from fatbulls.apis.api_fqdatac import get_daysrange
from rqalpha import run_file


class PortfolioBacktestEngine(AbstractBackTestEngine):

    def __init__(self, backtest_config: dict, strategy_file_path: str = None):
        self.backtest_config = backtest_config
        self.matching_type = backtest_config['mod']['sys_simulation']['matching_type']
        self.run_type = backtest_config['base']['run_type']
        self.start_date = self.backtest_config['base']['start_date']
        self.end_date = self.backtest_config['base']['end_date']
        self.strategy_file_path = strategy_file_path
        self.report_path = None
        self.trade_days = None
        self.portfolios = None

    def run(self):
        self.init_report_path()
        self.trade_days = get_daysrange(self.start_date, self.end_date)
        portfolio_dirpath = self.backtest_config['faithquant']['portfolio']['portfolio_dirpath']
        protfolio_name = portfolio_dirpath.rstrip("/").split("/")[-2]
        self.portfolios = Portfolios(protfolio_name, portfolio_dirpath)
        self.portfolios.check_portfolio(self.trade_days)
        self.backtest_config['faithquant']['portfolio']['stock_positions'] = self.portfolios.get_portfolio_data(self.start_date, self.end_date)
        run_file(self.strategy_file_path, config=self.backtest_config)
        self.save_success_flag()

    def save_success_flag(self):
        """保存report"""
        # 写入_SUCCESS
        if self.report_path is None:
            self.init_report_path()
        filename = self.report_path.rstrip("/") + "/../_SUCCESS"
        if os.path.isfile(filename):
            os.remove(filename)
        with open(filename, 'w+') as fp:
            fp.write("1")

    def init_report_path(self):
        try:
            base_report_path = self.backtest_config['faithquant']['portfolio']['base_report_path']
            task_num = self.backtest_config['faithquant']['portfolio']['task_num']
            self.report_path = os.path.join(base_report_path, task_num, 'data')
            self.backtest_config['mod']['sys_analyser']['record'] = True
            self.backtest_config['mod']['sys_analyser']['output_file'] = "{}/{}.pkl".format(self.report_path, task_num)
            self.backtest_config['mod']['sys_analyser']['report_save_path'] = self.report_path
            self.backtest_config['mod']['sys_analyser']['enabled'] = True
            self.backtest_config['mod']['sys_analyser']['plot'] = True
            self.backtest_config['mod']['sys_analyser']['plot_save_file'] = self.report_path
            os.makedirs(self.report_path, exist_ok=True)
        except Exception as e:
            raise FactorBackTestException("get_report_path error: {}".format(str(e)))
